#!/usr/bin/perl

#
# kgdb_demux
#
# This perl script attempts to separate console data from debugger  
# data, with the intent to make it look like two independent serial ports 
# (or at least that's the theory).  The default serial device is 
# /dev/ttyS0.
# 
# Usage: kgdb_demux [serial_device]
#
# Author: Brian Moyle
#         bmoyle@mvista.com
#
# Copyright 2000 MontaVista Software Inc.
#
#  This program is free software; you can redistribute  it and/or modify it
#  under  the terms of  the GNU General  Public License as published by the
#  Free Software Foundation;  either version 2 of the  License, or (at your
#  option) any later version.
#
#  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
#  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
#  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
#  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  You should have received a copy of the  GNU General Public License along
#  with this program; if not, write  to the Free Software Foundation, Inc.,
#  675 Mass Ave, Cambridge, MA 02139, USA.
#

# version: 1.01

# examples of serial data we'll be searching for:
#    $Z0,c00d61b4,4#3a
#    hel$c#63lo   { "hello" -> console, "$c#63" -> debug }

use blib;
use IO::Pty;
use IO::Select;

$serial_port = $ARGV[0] || "/dev/ttyS0";

print "\n";

# create pseudo-terminal for serial console
$console = new IO::Pty;
$console_slave = $console->slave;
print "console pty = ", $console_slave->ttyname, "\n";

# create pseudo-terminal for serial debugger
$debug = new IO::Pty;
$debug_slave  = $debug->slave;
print "debug pty   = ", $debug_slave->ttyname, "\n";

# open serial port
print "(serial port = $serial_port)\n";
open(SERIAL_IN, "< $serial_port") || die;
open(SERIAL_OUT, "> $serial_port") || die;

print "\n";

# watch for any input
$select = IO::Select->new();
$select->add(*SERIAL_IN);
$select->add($debug);
$select->add($console);

# maximum buffer size for debugging commands
$BUFMAX = 2048;

$found_start_char = "false";
while (1)  {
    # wait for one second, if nothing's pending
    if (@ready = $select->can_read(0.07))  {
        # cycle through filehandles reporting data available
        foreach (@ready)  {
            $bytes = sysread($_, $buffer, 1) || die;
            ### handle serial input ###
            if (fileno($_) eq fileno(SERIAL_IN))  {
                if ($found_start_char eq "true")  {
                    $count++;
                    # check whether we've exceeded the maximum count
                    if ($count >= $BUFMAX)  {
                        $count = 0;
                        $stored_buffer .= $buffer;
                        syswrite($console, $stored_buffer, 
                            length($stored_buffer));
                        $stored_buffer = "";
                        $found_start_char = "false";
                    }  else  {
                        # are we on the last two characters (checksum)?
                        if ($last_two > 0)  {
                            $last_two--;
                            $stored_buffer .= $buffer;
                            if ($last_two == 0)  {
                                @characters = split(//,$stored_buffer);
                                $checksum = 0;
                                foreach (@characters)  {
                                    $checksum += ord($_);
                                }
                                # remove '$', '#', and checksum from checksum
                                $checksum -= ord('$');
                                $checksum -= ord('#');
                                $checksum -= ord(@characters[$#characters-1]);
                                $checksum -= ord(@characters[$#characters]);
                                $checksum %= 256;
                                $hex_checksum = sprintf("%02x", $checksum);
                                $scanned_checksum =$characters[$#characters-1].
                                                   $characters[$#characters];
                                if ($hex_checksum eq $scanned_checksum)  {
                                    syswrite($debug, $stored_buffer, 
                                        length($stored_buffer));
                                }  else  {
                                    syswrite($console, $stored_buffer, 
                                        length($stored_buffer));
                                }
                                $stored_buffer = "";
                                $found_start_char = "false";
                            }
                        }  else  {
                            # if '#', we're ready for last two chars (count=2)
                            if ($buffer eq '#')  {
                                $last_two = 2;
                            }
                            if ($buffer eq '$')  {
                                # string was invalid -- print it and start over
                                syswrite($console, $stored_buffer, 
                                    length($stored_buffer));
                                $count = 0;
                                $stored_buffer = "";
                            }
                            $stored_buffer .= $buffer;
                        }
                    }
                }  else  {
                    if ($buffer eq '$')  {
                        # start searching for debug pattern
                        $count=1;
                        $stored_buffer = "";
                        $stored_buffer .= $buffer;
                        $found_start_char = "true";
                    }  else  {
                        # send to debugger if this was a response to a message
                        if (($response_pending eq "true") &&
                                (($buffer eq '+') || ($buffer eq '-')))  {
                            syswrite($debug, $buffer, length($buffer));
                            $response_pending = "false";
                        }  else  {
                            syswrite($console, $buffer, length($buffer));
                        }
                    }
                }
            }
            ### handle debugging input ###
            if (fileno($_) == fileno($debug))  {
                # this is only necessary to sort out responses in the console
                # (i.e., we don't need a similar flag in the other direction)
                $response_pending = "true";
                syswrite(SERIAL_OUT, $buffer, 1);
            }
            ### handle console input ###
            if (fileno($_) == fileno($console))  {
                syswrite(SERIAL_OUT, $buffer, 1);
            }
        }
    }  else  {
        # if we're in the middle of a debug cmd, we've waited too long -- flush
        if ($stored_buffer ne "")  {
            $count = 0;
            syswrite($console, $stored_buffer, 
                length($stored_buffer));
            $stored_buffer = "";
            $found_start_char = "false";
     
        }
    }
}

# close
close($console);
close($debug);
close(SERIAL_IN);
close(SERIAL_OUT);

